from __future__ import print_function, unicode_literals

import inspect
import re
import collections
import keyword
import renpy.sl2
import shutil
import io
import os
import textwrap

try:
    import builtins
except ImportError:
    import __builtin__ as builtins

# Additional keywords in the Ren'Py script language.
SCRIPT_KEYWORDS = """\
$
as
at
behind
expression
onlayer
zorder
strings
take
nointeract
elif
old
new
"""

# ATL Keywords.
ATL_KEYWORDS = """\
repeat
block
contains
parallel
choice
on
time
function
event
animation
clockwise
counterclockwise
circles
knot
"""

# SL2 Keywords (that aren't statements).
SL2_KEYWORDS = """\
tag
has
index
"""

# LayerImage keywords
LI_KEYWORDS = """\
layeredimage
attribute
group
always
"""


def script_keywords():

    tries = [ renpy.parser.statements ]

    rv = set()

    while tries:
        trie = tries.pop(0)
        for k, v in trie.words.items():
            rv.add(k)
            tries.append(v)

    rv.remove("layer")

    return rv


script_keywords()


def sl2_keywords():

    rv = set()

    for i in renpy.sl2.slparser.all_statements:
        rv.add(i.name)

    rv.remove("icon")
    rv.remove("iconbutton")

    return rv


def sl2_regexps():

    rv = [ ]

    groups = collections.defaultdict(set)
    has_style = { }

    for k, v in renpy.sl2.slparser.properties.items():
        prefix, style = k

        has_style[prefix] = style

        if prefix == "icon_":
            continue

        props = tuple(sorted(v))

        groups[props, style].add(prefix)

    style_part2 = "(?:" + "|".join(sorted(renpy.sl2.slparser.STYLE_PREFIXES)) + ")"

    items = list(groups.items())
    items.sort(key=lambda a : (tuple(sorted(a[1])), a[0][1]))

    for k, prefixes in items:
        names, style = k

        if len(prefixes) > 1:
            part1 = "(?:" + "|".join(prefixes) + ")"
        else:
            part1 = tuple(prefixes)[0]

        if style:
            part2 = style_part2
        else:
            part2 = ""

        part3 = "(?:" + "|".join(sorted(names)) + ")"

        re = part1 + part2 + part3
        rv.append(re)

    return rv


def expanded_sl2_properties():

    rv = set()

    for k, v in renpy.sl2.slparser.properties.items():
        prefix, style = k

        if prefix == "icon_":
            continue

        if style:
            style_prefixes = renpy.sl2.slparser.STYLE_PREFIXES
        else:
            style_prefixes = [ '' ]

        for i in style_prefixes:
            for j in v:
                rv.add(prefix + i + j)

    rv = list(rv)
    rv.sort()

    return rv


def write_keywords(srcdir='source'):
    outf = os.path.join(srcdir, 'keywords.py')

    kwlist = set(keyword.kwlist)
    kwlist |= script_keywords()
    kwlist |= sl2_keywords()
    kwlist |= set(ATL_KEYWORDS.split())
    kwlist |= set(SCRIPT_KEYWORDS.split())
    kwlist |= set(SL2_KEYWORDS.split())
    kwlist |= set(LI_KEYWORDS.split())

    kwlist = list(kwlist)

    kwlist.sort()

    with open(outf, "w") as f:

        f.write("keywords = %r\n" % kwlist)
        f.write("keyword_regex = %r\n" % ("|".join(re.escape(i) for i in kwlist)))

        properties = [ i for i in expanded_sl2_properties() if i not in kwlist ]

        f.write("properties = %r\n" % properties)

        f.write("property_regexes = %r\n" % sl2_regexps())

    shutil.copy(outf, os.path.join(srcdir, "../../tutorial/game/keywords.py"))


# A map from filename to a list of lines that are supposed to go into
# that file.
line_buffer = collections.defaultdict(list)

# A map from id(o) to the names it's documented under.
documented = collections.defaultdict(list)

# This keeps all objectsd we see alive, to prevent duplicates in documented.
documented_list = [ ]


def scan(name, o, prefix=""):

    doc_type = "function"

    # The section it's going into.
    section = None

    # The formatted arguments.
    args = None

    # Get the function's docstring.
    doc = inspect.getdoc(o)

    if not doc:
        return

    if doc[0] == ' ':
        print("Bad docstring for ", name, repr(doc))

    if re.match(r'[\w\.]+\(', doc):
        doc = doc.partition("\n\n")[2]
        doc = textwrap.dedent(doc)

    # Break up the doc string, scan it for specials.
    lines = [ ]

    for l in doc.split("\n"):

        m = re.match(r':doc: *(\w+) *(\w+)?', l)
        if m:
            section = m.group(1)

            if m.group(2):
                doc_type = m.group(2)

            continue

        m = re.match(r':args: *(.*)', l)
        if m:
            args = m.group(1)
            continue

        m = re.match(r':name: *(\S+)', l)
        if m:
            if name != m.group(1):
                return
            continue

        lines.append(l)

    if section is None:
        return

    if args is None:

        # Get the arguments.
        if inspect.isclass(o):
            init = getattr(o, "__init__", None)
            if not init:
                return

            init_doc = inspect.getdoc(init)

            if init_doc and not init_doc.startswith("x.__init__("):
                lines.append("")
                lines.extend(init_doc.split("\n"))

            try:
                args = inspect.getargspec(init)
            except Exception:
                args = None

        elif inspect.isfunction(o):
            args = inspect.getargspec(o)

        elif inspect.ismethod(o):
            args = inspect.getargspec(o)

        else:
            print("Warning: %s has section but not args." % name)

            return

        # Format the arguments.
        if args is not None:

            args = inspect.formatargspec(*args)
            args = args.replace("(self, ", "(")
        else:
            args = "()"

    # Put it into the line buffer.
    lb = line_buffer[section]

    lb.append(prefix + ".. %s:: %s%s" % (doc_type, name, args))

    for l in lines:
        lb.append(prefix + "    " + l)

    lb.append(prefix + "")

    if inspect.isclass(o):
        if (name not in [ "Matrix", "OffsetMatrix", "RotateMatrix", "ScaleMatrix" ]):
            for i in dir(o):
                scan(i, getattr(o, i), prefix + "    ")

    if name == "identity":
        raise Exception("identity")

    documented_list.append(o)
    documented[id(o)].append(name)


def scan_section(name, o):
    """
    Scans object o. Assumes it has the name name.
    """

    for n in dir(o):
        scan(name + n, getattr(o, n))


def write_line_buffer(incdir='source/inc'):

    for k, v in line_buffer.items():

        f = io.StringIO()

        print(u".. Automatically generated file - do not modify.", file=f)
        print(u"", file=f)

        for l in v:
            print(l, file=f)

        s = f.getvalue()

        fname = os.path.join(incdir, k)
        if os.path.exists(fname):
            with open(fname) as f:
                if f.read() == s:
                    print("Retaining", k)
                    continue

        print("Generating", k)

        with open(fname, "w") as f:
            f.write(s)


name_kind = collections.defaultdict(str)


def scan_docs(srcdir='source', incdir='source/inc'):
    """
    Scans the documentation for functions, classes, and variables.
    """

    def scan_file(fn):
        f = open(fn)

        for l in f:
            m = re.search(r"\.\. (\w+):: ([.\w+]+)", l)

            if not m:
                continue

            name_kind[m.group(2)] = m.group(1)

    for i in os.listdir(srcdir):
        if i.endswith(".rst"):
            scan_file(os.path.join(srcdir, i))

    for i in os.listdir(incdir):
        scan_file(os.path.join(incdir, i))


def format_name(name):

    if name_kind[name] == 'function':
        name = ":func:`{}`".format(name)
    elif name_kind[name] == 'class':
        name = ":class:`{}`".format(name)
    elif name_kind[name] == 'var':
        name = ":var:`{}`".format(name)

    return name


def write_reserved(module, dest, ignore_builtins):

    with open(dest, "w") as f:

        for i in sorted(dir(module)):

            if i == "doc":
                continue

            if i.startswith("_"):
                continue

            if ignore_builtins and hasattr(builtins, i):
                continue

            f.write("* " + format_name(i) + "\n")


def write_pure_const(incdir='source/inc'):

    def write_set(f, s):
        l = list(s)
        l.sort()

        for i in l:
            f.write("* " + format_name(i) + "\n")

    pure = renpy.pyanalysis.pure_functions # @UndefinedVariable
    constants = renpy.pyanalysis.constants - pure # @UndefinedVariable

    with open(os.path.join(incdir, "pure_vars"), "w") as f:
        write_set(f, pure)

    with open(os.path.join(incdir, "const_vars"), "w") as f:
        write_set(f, constants)


def write_easings(ns, incdir='source/inc'):

    with open(os.path.join(incdir, "easings"), "w") as f:
        f.write(".. csv-table::\n")
        f.write('    :header: "Ren\'Py Name", "easings.net Name"\n')
        f.write('\n')

        for name in sorted(dir(ns)):
            if not name.startswith("ease"):
                continue
            if not "_" in name:
                continue

            stdname = name.replace("ease_", "easeInOut_").replace("easein_", "easeOut_").replace("easeout_", "easeIn_")
            f.write('    "{}", "{}"\n'.format(name, stdname))


def tq_script(name, srcdir='source'):

    with open(os.path.join(srcdir, "../../the_question/game", name), "r") as f:
        lines = f.readlines()
        lines = [ ("    " + i).rstrip() for i in lines ]
        return "\n".join(lines)


def write_tq(srcdir='source'):
    script = tq_script("script.rpy", srcdir=srcdir)
    options = tq_script("options.rpy", srcdir=srcdir)

    with open(os.path.join(srcdir, "thequestion.txt"), "r") as f:
        template = f.read()

    with open(os.path.join(srcdir, "thequestion.rst"), "w") as f:
        f.write(template.format(script=script, options=options))


def check_dups():

    duplicates = False

    for v in documented.values():
        if len(v) >= 2:
            duplicates = True

            print(" and ".join(v), "are duplicate.")

    if duplicates:
        raise SystemExit(1)



from sdl2 cimport SDL_GL_GetProcAddress, SDL_GetError

found_functions = set()

cdef void *find_gl_command(names):

    cdef void *rv = NULL

    for i in names:
        rv = SDL_GL_GetProcAddress(i)

        if rv != NULL:
            found_functions.add(names[0].decode("utf-8"))
            return rv

    import renpy
    renpy.display.log.write("UGUU couldn't find {}: {}".format(names[0], SDL_GetError()))

    return NULL

def clear_missing_functions():
    global found_functions
    found_functions = set()

def check_missing_functions(required):
    import renpy

    required = set(required)

    missing_required = list(required - found_functions)
    missing_required.sort()

    if missing_required:

        renpy.display.log.write("The following gl functions are missing:")
        for i in missing_required:
            renpy.display.log.write("- %s", i)

        return True
    else:
        return False



cdef glActiveTexture_type glActiveTexture


cdef glAttachShader_type glAttachShader


cdef glBeginQuery_type glBeginQuery


cdef glBeginTransformFeedback_type glBeginTransformFeedback


cdef glBindAttribLocation_type glBindAttribLocation


cdef glBindBuffer_type glBindBuffer


cdef glBindBufferBase_type glBindBufferBase


cdef glBindBufferRange_type glBindBufferRange


cdef glBindFramebuffer_type glBindFramebuffer


cdef glBindRenderbuffer_type glBindRenderbuffer


cdef glBindTexture_type glBindTexture


cdef glBindVertexArray_type glBindVertexArray


cdef glBlendColor_type glBlendColor


cdef glBlendEquation_type glBlendEquation


cdef glBlendEquationSeparate_type glBlendEquationSeparate


cdef glBlendFunc_type glBlendFunc


cdef glBlendFuncSeparate_type glBlendFuncSeparate


cdef glBlitFramebuffer_type glBlitFramebuffer


cdef glBufferData_type glBufferData


cdef glBufferSubData_type glBufferSubData


cdef glCheckFramebufferStatus_type glCheckFramebufferStatus


cdef glClear_type glClear


cdef glClearBufferfi_type glClearBufferfi


cdef glClearBufferfv_type glClearBufferfv


cdef glClearBufferiv_type glClearBufferiv


cdef glClearBufferuiv_type glClearBufferuiv


cdef glClearColor_type glClearColor


cdef glClearStencil_type glClearStencil


cdef glColorMask_type glColorMask


cdef glCompileShader_type glCompileShader


cdef glCompressedTexImage2D_type glCompressedTexImage2D


cdef glCompressedTexImage3D_type glCompressedTexImage3D


cdef glCompressedTexSubImage2D_type glCompressedTexSubImage2D


cdef glCompressedTexSubImage3D_type glCompressedTexSubImage3D


cdef glCopyTexImage2D_type glCopyTexImage2D


cdef glCopyTexSubImage2D_type glCopyTexSubImage2D


cdef glCopyTexSubImage3D_type glCopyTexSubImage3D


cdef glCreateProgram_type glCreateProgram


cdef glCreateShader_type glCreateShader


cdef glCullFace_type glCullFace


cdef glDeleteBuffers_type glDeleteBuffers


cdef glDeleteFramebuffers_type glDeleteFramebuffers


cdef glDeleteProgram_type glDeleteProgram


cdef glDeleteQueries_type glDeleteQueries


cdef glDeleteRenderbuffers_type glDeleteRenderbuffers


cdef glDeleteShader_type glDeleteShader


cdef glDeleteTextures_type glDeleteTextures


cdef glDeleteVertexArrays_type glDeleteVertexArrays


cdef glDepthFunc_type glDepthFunc


cdef glDepthMask_type glDepthMask


cdef glDetachShader_type glDetachShader


cdef glDisable_type glDisable


cdef glDisableVertexAttribArray_type glDisableVertexAttribArray


cdef glDrawArrays_type glDrawArrays


cdef glDrawBuffers_type glDrawBuffers


cdef glDrawElements_type glDrawElements


cdef glDrawRangeElements_type glDrawRangeElements


cdef glEnable_type glEnable


cdef glEnableVertexAttribArray_type glEnableVertexAttribArray


cdef glEndQuery_type glEndQuery


cdef glEndTransformFeedback_type glEndTransformFeedback


cdef glFinish_type glFinish


cdef glFlush_type glFlush


cdef glFlushMappedBufferRange_type glFlushMappedBufferRange


cdef glFramebufferRenderbuffer_type glFramebufferRenderbuffer


cdef glFramebufferTexture2D_type glFramebufferTexture2D


cdef glFramebufferTextureLayer_type glFramebufferTextureLayer


cdef glFrontFace_type glFrontFace


cdef glGenBuffers_type glGenBuffers


cdef glGenFramebuffers_type glGenFramebuffers


cdef glGenQueries_type glGenQueries


cdef glGenRenderbuffers_type glGenRenderbuffers


cdef glGenTextures_type glGenTextures


cdef glGenVertexArrays_type glGenVertexArrays


cdef glGenerateMipmap_type glGenerateMipmap


cdef glGetActiveAttrib_type glGetActiveAttrib


cdef glGetActiveUniform_type glGetActiveUniform


cdef glGetAttachedShaders_type glGetAttachedShaders


cdef glGetAttribLocation_type glGetAttribLocation


cdef glGetBooleanv_type glGetBooleanv


cdef glGetBufferParameteriv_type glGetBufferParameteriv


cdef glGetBufferPointerv_type glGetBufferPointerv


cdef glGetError_type glGetError


cdef glGetFloatv_type glGetFloatv


cdef glGetFragDataLocation_type glGetFragDataLocation


cdef glGetFramebufferAttachmentParameteriv_type glGetFramebufferAttachmentParameteriv


cdef glGetIntegeri_v_type glGetIntegeri_v


cdef glGetIntegerv_type glGetIntegerv


cdef glGetProgramInfoLog_type glGetProgramInfoLog


cdef glGetProgramiv_type glGetProgramiv


cdef glGetQueryObjectuiv_type glGetQueryObjectuiv


cdef glGetQueryiv_type glGetQueryiv


cdef glGetRenderbufferParameteriv_type glGetRenderbufferParameteriv


cdef glGetShaderInfoLog_type glGetShaderInfoLog


cdef glGetShaderSource_type glGetShaderSource


cdef glGetShaderiv_type glGetShaderiv


cdef glGetString_type glGetString


cdef glGetStringi_type glGetStringi


cdef glGetTexParameterfv_type glGetTexParameterfv


cdef glGetTexParameteriv_type glGetTexParameteriv


cdef glGetTransformFeedbackVarying_type glGetTransformFeedbackVarying


cdef glGetUniformLocation_type glGetUniformLocation


cdef glGetUniformfv_type glGetUniformfv


cdef glGetUniformiv_type glGetUniformiv


cdef glGetUniformuiv_type glGetUniformuiv


cdef glGetVertexAttribIiv_type glGetVertexAttribIiv


cdef glGetVertexAttribIuiv_type glGetVertexAttribIuiv


cdef glGetVertexAttribPointerv_type glGetVertexAttribPointerv


cdef glGetVertexAttribfv_type glGetVertexAttribfv


cdef glGetVertexAttribiv_type glGetVertexAttribiv


cdef glHint_type glHint


cdef glIsBuffer_type glIsBuffer


cdef glIsEnabled_type glIsEnabled


cdef glIsFramebuffer_type glIsFramebuffer


cdef glIsProgram_type glIsProgram


cdef glIsQuery_type glIsQuery


cdef glIsRenderbuffer_type glIsRenderbuffer


cdef glIsShader_type glIsShader


cdef glIsTexture_type glIsTexture


cdef glIsVertexArray_type glIsVertexArray


cdef glLineWidth_type glLineWidth


cdef glLinkProgram_type glLinkProgram


cdef glMapBufferRange_type glMapBufferRange


cdef glPixelStorei_type glPixelStorei


cdef glPolygonOffset_type glPolygonOffset


cdef glReadBuffer_type glReadBuffer


cdef glReadPixels_type glReadPixels


cdef glRenderbufferStorage_type glRenderbufferStorage


cdef glRenderbufferStorageMultisample_type glRenderbufferStorageMultisample


cdef glSampleCoverage_type glSampleCoverage


cdef glScissor_type glScissor


cdef glShaderSource_type glShaderSource


cdef glStencilFunc_type glStencilFunc


cdef glStencilFuncSeparate_type glStencilFuncSeparate


cdef glStencilMask_type glStencilMask


cdef glStencilMaskSeparate_type glStencilMaskSeparate


cdef glStencilOp_type glStencilOp


cdef glStencilOpSeparate_type glStencilOpSeparate


cdef glTexImage2D_type glTexImage2D


cdef glTexImage3D_type glTexImage3D


cdef glTexParameterf_type glTexParameterf


cdef glTexParameterfv_type glTexParameterfv


cdef glTexParameteri_type glTexParameteri


cdef glTexParameteriv_type glTexParameteriv


cdef glTexSubImage2D_type glTexSubImage2D


cdef glTexSubImage3D_type glTexSubImage3D


cdef glTransformFeedbackVaryings_type glTransformFeedbackVaryings


cdef glUniform1f_type glUniform1f


cdef glUniform1fv_type glUniform1fv


cdef glUniform1i_type glUniform1i


cdef glUniform1iv_type glUniform1iv


cdef glUniform1ui_type glUniform1ui


cdef glUniform1uiv_type glUniform1uiv


cdef glUniform2f_type glUniform2f


cdef glUniform2fv_type glUniform2fv


cdef glUniform2i_type glUniform2i


cdef glUniform2iv_type glUniform2iv


cdef glUniform2ui_type glUniform2ui


cdef glUniform2uiv_type glUniform2uiv


cdef glUniform3f_type glUniform3f


cdef glUniform3fv_type glUniform3fv


cdef glUniform3i_type glUniform3i


cdef glUniform3iv_type glUniform3iv


cdef glUniform3ui_type glUniform3ui


cdef glUniform3uiv_type glUniform3uiv


cdef glUniform4f_type glUniform4f


cdef glUniform4fv_type glUniform4fv


cdef glUniform4i_type glUniform4i


cdef glUniform4iv_type glUniform4iv


cdef glUniform4ui_type glUniform4ui


cdef glUniform4uiv_type glUniform4uiv


cdef glUniformMatrix2fv_type glUniformMatrix2fv


cdef glUniformMatrix2x3fv_type glUniformMatrix2x3fv


cdef glUniformMatrix2x4fv_type glUniformMatrix2x4fv


cdef glUniformMatrix3fv_type glUniformMatrix3fv


cdef glUniformMatrix3x2fv_type glUniformMatrix3x2fv


cdef glUniformMatrix3x4fv_type glUniformMatrix3x4fv


cdef glUniformMatrix4fv_type glUniformMatrix4fv


cdef glUniformMatrix4x2fv_type glUniformMatrix4x2fv


cdef glUniformMatrix4x3fv_type glUniformMatrix4x3fv


cdef glUnmapBuffer_type glUnmapBuffer


cdef glUseProgram_type glUseProgram


cdef glValidateProgram_type glValidateProgram


cdef glVertexAttrib1f_type glVertexAttrib1f


cdef glVertexAttrib1fv_type glVertexAttrib1fv


cdef glVertexAttrib2f_type glVertexAttrib2f


cdef glVertexAttrib2fv_type glVertexAttrib2fv


cdef glVertexAttrib3f_type glVertexAttrib3f


cdef glVertexAttrib3fv_type glVertexAttrib3fv


cdef glVertexAttrib4f_type glVertexAttrib4f


cdef glVertexAttrib4fv_type glVertexAttrib4fv


cdef glVertexAttribI4i_type glVertexAttribI4i


cdef glVertexAttribI4iv_type glVertexAttribI4iv


cdef glVertexAttribI4ui_type glVertexAttribI4ui


cdef glVertexAttribI4uiv_type glVertexAttribI4uiv


cdef glVertexAttribIPointer_type glVertexAttribIPointer


cdef glVertexAttribPointer_type glVertexAttribPointer


cdef glViewport_type glViewport


def load():

    global glActiveTexture
    glActiveTexture = <glActiveTexture_type> find_gl_command([b'glActiveTexture', b'glActiveTextureARB'])

    global glAttachShader
    glAttachShader = <glAttachShader_type> find_gl_command([b'glAttachShader'])

    global glBeginQuery
    glBeginQuery = <glBeginQuery_type> find_gl_command([b'glBeginQuery', b'glBeginQueryARB'])

    global glBeginTransformFeedback
    glBeginTransformFeedback = <glBeginTransformFeedback_type> find_gl_command([b'glBeginTransformFeedback', b'glBeginTransformFeedbackEXT', b'glBeginTransformFeedbackNV'])

    global glBindAttribLocation
    glBindAttribLocation = <glBindAttribLocation_type> find_gl_command([b'glBindAttribLocation', b'glBindAttribLocationARB'])

    global glBindBuffer
    glBindBuffer = <glBindBuffer_type> find_gl_command([b'glBindBuffer', b'glBindBufferARB'])

    global glBindBufferBase
    glBindBufferBase = <glBindBufferBase_type> find_gl_command([b'glBindBufferBase', b'glBindBufferBaseEXT', b'glBindBufferBaseNV'])

    global glBindBufferRange
    glBindBufferRange = <glBindBufferRange_type> find_gl_command([b'glBindBufferRange', b'glBindBufferRangeEXT', b'glBindBufferRangeNV'])

    global glBindFramebuffer
    glBindFramebuffer = <glBindFramebuffer_type> find_gl_command([b'glBindFramebuffer', b'glBindFramebufferEXT'])

    global glBindRenderbuffer
    glBindRenderbuffer = <glBindRenderbuffer_type> find_gl_command([b'glBindRenderbuffer', b'glBindRenderbufferEXT'])

    global glBindTexture
    glBindTexture = <glBindTexture_type> find_gl_command([b'glBindTexture', b'glBindTextureEXT'])

    global glBindVertexArray
    glBindVertexArray = <glBindVertexArray_type> find_gl_command([b'glBindVertexArray', b'glBindVertexArrayOES'])

    global glBlendColor
    glBlendColor = <glBlendColor_type> find_gl_command([b'glBlendColor', b'glBlendColorEXT'])

    global glBlendEquation
    glBlendEquation = <glBlendEquation_type> find_gl_command([b'glBlendEquation', b'glBlendEquationEXT'])

    global glBlendEquationSeparate
    glBlendEquationSeparate = <glBlendEquationSeparate_type> find_gl_command([b'glBlendEquationSeparate', b'glBlendEquationSeparateEXT'])

    global glBlendFunc
    glBlendFunc = <glBlendFunc_type> find_gl_command([b'glBlendFunc'])

    global glBlendFuncSeparate
    glBlendFuncSeparate = <glBlendFuncSeparate_type> find_gl_command([b'glBlendFuncSeparate', b'glBlendFuncSeparateEXT', b'glBlendFuncSeparateINGR'])

    global glBlitFramebuffer
    glBlitFramebuffer = <glBlitFramebuffer_type> find_gl_command([b'glBlitFramebuffer', b'glBlitFramebufferEXT', b'glBlitFramebufferNV'])

    global glBufferData
    glBufferData = <glBufferData_type> find_gl_command([b'glBufferData', b'glBufferDataARB'])

    global glBufferSubData
    glBufferSubData = <glBufferSubData_type> find_gl_command([b'glBufferSubData', b'glBufferSubDataARB'])

    global glCheckFramebufferStatus
    glCheckFramebufferStatus = <glCheckFramebufferStatus_type> find_gl_command([b'glCheckFramebufferStatus', b'glCheckFramebufferStatusEXT'])

    global glClear
    glClear = <glClear_type> find_gl_command([b'glClear'])

    global glClearBufferfi
    glClearBufferfi = <glClearBufferfi_type> find_gl_command([b'glClearBufferfi'])

    global glClearBufferfv
    glClearBufferfv = <glClearBufferfv_type> find_gl_command([b'glClearBufferfv'])

    global glClearBufferiv
    glClearBufferiv = <glClearBufferiv_type> find_gl_command([b'glClearBufferiv'])

    global glClearBufferuiv
    glClearBufferuiv = <glClearBufferuiv_type> find_gl_command([b'glClearBufferuiv'])

    global glClearColor
    glClearColor = <glClearColor_type> find_gl_command([b'glClearColor'])

    global glClearStencil
    glClearStencil = <glClearStencil_type> find_gl_command([b'glClearStencil'])

    global glColorMask
    glColorMask = <glColorMask_type> find_gl_command([b'glColorMask'])

    global glCompileShader
    glCompileShader = <glCompileShader_type> find_gl_command([b'glCompileShader', b'glCompileShaderARB'])

    global glCompressedTexImage2D
    glCompressedTexImage2D = <glCompressedTexImage2D_type> find_gl_command([b'glCompressedTexImage2D', b'glCompressedTexImage2DARB'])

    global glCompressedTexImage3D
    glCompressedTexImage3D = <glCompressedTexImage3D_type> find_gl_command([b'glCompressedTexImage3D', b'glCompressedTexImage3DARB'])

    global glCompressedTexSubImage2D
    glCompressedTexSubImage2D = <glCompressedTexSubImage2D_type> find_gl_command([b'glCompressedTexSubImage2D', b'glCompressedTexSubImage2DARB'])

    global glCompressedTexSubImage3D
    glCompressedTexSubImage3D = <glCompressedTexSubImage3D_type> find_gl_command([b'glCompressedTexSubImage3D', b'glCompressedTexSubImage3DARB'])

    global glCopyTexImage2D
    glCopyTexImage2D = <glCopyTexImage2D_type> find_gl_command([b'glCopyTexImage2D', b'glCopyTexImage2DEXT'])

    global glCopyTexSubImage2D
    glCopyTexSubImage2D = <glCopyTexSubImage2D_type> find_gl_command([b'glCopyTexSubImage2D', b'glCopyTexSubImage2DEXT'])

    global glCopyTexSubImage3D
    glCopyTexSubImage3D = <glCopyTexSubImage3D_type> find_gl_command([b'glCopyTexSubImage3D', b'glCopyTexSubImage3DEXT'])

    global glCreateProgram
    glCreateProgram = <glCreateProgram_type> find_gl_command([b'glCreateProgram', b'glCreateProgramObjectARB'])

    global glCreateShader
    glCreateShader = <glCreateShader_type> find_gl_command([b'glCreateShader', b'glCreateShaderObjectARB'])

    global glCullFace
    glCullFace = <glCullFace_type> find_gl_command([b'glCullFace'])

    global glDeleteBuffers
    glDeleteBuffers = <glDeleteBuffers_type> find_gl_command([b'glDeleteBuffers', b'glDeleteBuffersARB'])

    global glDeleteFramebuffers
    glDeleteFramebuffers = <glDeleteFramebuffers_type> find_gl_command([b'glDeleteFramebuffers', b'glDeleteFramebuffersEXT'])

    global glDeleteProgram
    glDeleteProgram = <glDeleteProgram_type> find_gl_command([b'glDeleteProgram'])

    global glDeleteQueries
    glDeleteQueries = <glDeleteQueries_type> find_gl_command([b'glDeleteQueries', b'glDeleteQueriesARB'])

    global glDeleteRenderbuffers
    glDeleteRenderbuffers = <glDeleteRenderbuffers_type> find_gl_command([b'glDeleteRenderbuffers', b'glDeleteRenderbuffersEXT'])

    global glDeleteShader
    glDeleteShader = <glDeleteShader_type> find_gl_command([b'glDeleteShader'])

    global glDeleteTextures
    glDeleteTextures = <glDeleteTextures_type> find_gl_command([b'glDeleteTextures'])

    global glDeleteVertexArrays
    glDeleteVertexArrays = <glDeleteVertexArrays_type> find_gl_command([b'glDeleteVertexArrays', b'glDeleteVertexArraysAPPLE', b'glDeleteVertexArraysOES'])

    global glDepthFunc
    glDepthFunc = <glDepthFunc_type> find_gl_command([b'glDepthFunc'])

    global glDepthMask
    glDepthMask = <glDepthMask_type> find_gl_command([b'glDepthMask'])

    global glDetachShader
    glDetachShader = <glDetachShader_type> find_gl_command([b'glDetachShader'])

    global glDisable
    glDisable = <glDisable_type> find_gl_command([b'glDisable'])

    global glDisableVertexAttribArray
    glDisableVertexAttribArray = <glDisableVertexAttribArray_type> find_gl_command([b'glDisableVertexAttribArray', b'glDisableVertexAttribArrayARB'])

    global glDrawArrays
    glDrawArrays = <glDrawArrays_type> find_gl_command([b'glDrawArrays', b'glDrawArraysEXT'])

    global glDrawBuffers
    glDrawBuffers = <glDrawBuffers_type> find_gl_command([b'glDrawBuffers', b'glDrawBuffersARB', b'glDrawBuffersATI', b'glDrawBuffersEXT'])

    global glDrawElements
    glDrawElements = <glDrawElements_type> find_gl_command([b'glDrawElements'])

    global glDrawRangeElements
    glDrawRangeElements = <glDrawRangeElements_type> find_gl_command([b'glDrawRangeElements', b'glDrawRangeElementsEXT'])

    global glEnable
    glEnable = <glEnable_type> find_gl_command([b'glEnable'])

    global glEnableVertexAttribArray
    glEnableVertexAttribArray = <glEnableVertexAttribArray_type> find_gl_command([b'glEnableVertexAttribArray', b'glEnableVertexAttribArrayARB'])

    global glEndQuery
    glEndQuery = <glEndQuery_type> find_gl_command([b'glEndQuery', b'glEndQueryARB'])

    global glEndTransformFeedback
    glEndTransformFeedback = <glEndTransformFeedback_type> find_gl_command([b'glEndTransformFeedback', b'glEndTransformFeedbackEXT', b'glEndTransformFeedbackNV'])

    global glFinish
    glFinish = <glFinish_type> find_gl_command([b'glFinish'])

    global glFlush
    glFlush = <glFlush_type> find_gl_command([b'glFlush'])

    global glFlushMappedBufferRange
    glFlushMappedBufferRange = <glFlushMappedBufferRange_type> find_gl_command([b'glFlushMappedBufferRange', b'glFlushMappedBufferRangeAPPLE', b'glFlushMappedBufferRangeEXT'])

    global glFramebufferRenderbuffer
    glFramebufferRenderbuffer = <glFramebufferRenderbuffer_type> find_gl_command([b'glFramebufferRenderbuffer', b'glFramebufferRenderbufferEXT'])

    global glFramebufferTexture2D
    glFramebufferTexture2D = <glFramebufferTexture2D_type> find_gl_command([b'glFramebufferTexture2D', b'glFramebufferTexture2DEXT'])

    global glFramebufferTextureLayer
    glFramebufferTextureLayer = <glFramebufferTextureLayer_type> find_gl_command([b'glFramebufferTextureLayer', b'glFramebufferTextureLayerARB', b'glFramebufferTextureLayerEXT'])

    global glFrontFace
    glFrontFace = <glFrontFace_type> find_gl_command([b'glFrontFace'])

    global glGenBuffers
    glGenBuffers = <glGenBuffers_type> find_gl_command([b'glGenBuffers', b'glGenBuffersARB'])

    global glGenFramebuffers
    glGenFramebuffers = <glGenFramebuffers_type> find_gl_command([b'glGenFramebuffers', b'glGenFramebuffersEXT'])

    global glGenQueries
    glGenQueries = <glGenQueries_type> find_gl_command([b'glGenQueries', b'glGenQueriesARB'])

    global glGenRenderbuffers
    glGenRenderbuffers = <glGenRenderbuffers_type> find_gl_command([b'glGenRenderbuffers', b'glGenRenderbuffersEXT'])

    global glGenTextures
    glGenTextures = <glGenTextures_type> find_gl_command([b'glGenTextures'])

    global glGenVertexArrays
    glGenVertexArrays = <glGenVertexArrays_type> find_gl_command([b'glGenVertexArrays', b'glGenVertexArraysAPPLE', b'glGenVertexArraysOES'])

    global glGenerateMipmap
    glGenerateMipmap = <glGenerateMipmap_type> find_gl_command([b'glGenerateMipmap', b'glGenerateMipmapEXT'])

    global glGetActiveAttrib
    glGetActiveAttrib = <glGetActiveAttrib_type> find_gl_command([b'glGetActiveAttrib', b'glGetActiveAttribARB'])

    global glGetActiveUniform
    glGetActiveUniform = <glGetActiveUniform_type> find_gl_command([b'glGetActiveUniform', b'glGetActiveUniformARB'])

    global glGetAttachedShaders
    glGetAttachedShaders = <glGetAttachedShaders_type> find_gl_command([b'glGetAttachedShaders'])

    global glGetAttribLocation
    glGetAttribLocation = <glGetAttribLocation_type> find_gl_command([b'glGetAttribLocation', b'glGetAttribLocationARB'])

    global glGetBooleanv
    glGetBooleanv = <glGetBooleanv_type> find_gl_command([b'glGetBooleanv'])

    global glGetBufferParameteriv
    glGetBufferParameteriv = <glGetBufferParameteriv_type> find_gl_command([b'glGetBufferParameteriv', b'glGetBufferParameterivARB'])

    global glGetBufferPointerv
    glGetBufferPointerv = <glGetBufferPointerv_type> find_gl_command([b'glGetBufferPointerv', b'glGetBufferPointervARB', b'glGetBufferPointervOES'])

    global glGetError
    glGetError = <glGetError_type> find_gl_command([b'glGetError'])

    global glGetFloatv
    glGetFloatv = <glGetFloatv_type> find_gl_command([b'glGetFloatv'])

    global glGetFragDataLocation
    glGetFragDataLocation = <glGetFragDataLocation_type> find_gl_command([b'glGetFragDataLocation', b'glGetFragDataLocationEXT'])

    global glGetFramebufferAttachmentParameteriv
    glGetFramebufferAttachmentParameteriv = <glGetFramebufferAttachmentParameteriv_type> find_gl_command([b'glGetFramebufferAttachmentParameteriv', b'glGetFramebufferAttachmentParameterivEXT'])

    global glGetIntegeri_v
    glGetIntegeri_v = <glGetIntegeri_v_type> find_gl_command([b'glGetIntegeri_v', b'glGetIntegerIndexedvEXT'])

    global glGetIntegerv
    glGetIntegerv = <glGetIntegerv_type> find_gl_command([b'glGetIntegerv'])

    global glGetProgramInfoLog
    glGetProgramInfoLog = <glGetProgramInfoLog_type> find_gl_command([b'glGetProgramInfoLog'])

    global glGetProgramiv
    glGetProgramiv = <glGetProgramiv_type> find_gl_command([b'glGetProgramiv'])

    global glGetQueryObjectuiv
    glGetQueryObjectuiv = <glGetQueryObjectuiv_type> find_gl_command([b'glGetQueryObjectuiv', b'glGetQueryObjectuivARB'])

    global glGetQueryiv
    glGetQueryiv = <glGetQueryiv_type> find_gl_command([b'glGetQueryiv', b'glGetQueryivARB'])

    global glGetRenderbufferParameteriv
    glGetRenderbufferParameteriv = <glGetRenderbufferParameteriv_type> find_gl_command([b'glGetRenderbufferParameteriv', b'glGetRenderbufferParameterivEXT'])

    global glGetShaderInfoLog
    glGetShaderInfoLog = <glGetShaderInfoLog_type> find_gl_command([b'glGetShaderInfoLog'])

    global glGetShaderSource
    glGetShaderSource = <glGetShaderSource_type> find_gl_command([b'glGetShaderSource', b'glGetShaderSourceARB'])

    global glGetShaderiv
    glGetShaderiv = <glGetShaderiv_type> find_gl_command([b'glGetShaderiv'])

    global glGetString
    glGetString = <glGetString_type> find_gl_command([b'glGetString'])

    global glGetStringi
    glGetStringi = <glGetStringi_type> find_gl_command([b'glGetStringi'])

    global glGetTexParameterfv
    glGetTexParameterfv = <glGetTexParameterfv_type> find_gl_command([b'glGetTexParameterfv'])

    global glGetTexParameteriv
    glGetTexParameteriv = <glGetTexParameteriv_type> find_gl_command([b'glGetTexParameteriv'])

    global glGetTransformFeedbackVarying
    glGetTransformFeedbackVarying = <glGetTransformFeedbackVarying_type> find_gl_command([b'glGetTransformFeedbackVarying', b'glGetTransformFeedbackVaryingEXT'])

    global glGetUniformLocation
    glGetUniformLocation = <glGetUniformLocation_type> find_gl_command([b'glGetUniformLocation', b'glGetUniformLocationARB'])

    global glGetUniformfv
    glGetUniformfv = <glGetUniformfv_type> find_gl_command([b'glGetUniformfv', b'glGetUniformfvARB'])

    global glGetUniformiv
    glGetUniformiv = <glGetUniformiv_type> find_gl_command([b'glGetUniformiv', b'glGetUniformivARB'])

    global glGetUniformuiv
    glGetUniformuiv = <glGetUniformuiv_type> find_gl_command([b'glGetUniformuiv', b'glGetUniformuivEXT'])

    global glGetVertexAttribIiv
    glGetVertexAttribIiv = <glGetVertexAttribIiv_type> find_gl_command([b'glGetVertexAttribIiv', b'glGetVertexAttribIivEXT'])

    global glGetVertexAttribIuiv
    glGetVertexAttribIuiv = <glGetVertexAttribIuiv_type> find_gl_command([b'glGetVertexAttribIuiv', b'glGetVertexAttribIuivEXT'])

    global glGetVertexAttribPointerv
    glGetVertexAttribPointerv = <glGetVertexAttribPointerv_type> find_gl_command([b'glGetVertexAttribPointerv', b'glGetVertexAttribPointervARB', b'glGetVertexAttribPointervNV'])

    global glGetVertexAttribfv
    glGetVertexAttribfv = <glGetVertexAttribfv_type> find_gl_command([b'glGetVertexAttribfv', b'glGetVertexAttribfvARB', b'glGetVertexAttribfvNV'])

    global glGetVertexAttribiv
    glGetVertexAttribiv = <glGetVertexAttribiv_type> find_gl_command([b'glGetVertexAttribiv', b'glGetVertexAttribivARB', b'glGetVertexAttribivNV'])

    global glHint
    glHint = <glHint_type> find_gl_command([b'glHint'])

    global glIsBuffer
    glIsBuffer = <glIsBuffer_type> find_gl_command([b'glIsBuffer', b'glIsBufferARB'])

    global glIsEnabled
    glIsEnabled = <glIsEnabled_type> find_gl_command([b'glIsEnabled'])

    global glIsFramebuffer
    glIsFramebuffer = <glIsFramebuffer_type> find_gl_command([b'glIsFramebuffer', b'glIsFramebufferEXT'])

    global glIsProgram
    glIsProgram = <glIsProgram_type> find_gl_command([b'glIsProgram'])

    global glIsQuery
    glIsQuery = <glIsQuery_type> find_gl_command([b'glIsQuery', b'glIsQueryARB'])

    global glIsRenderbuffer
    glIsRenderbuffer = <glIsRenderbuffer_type> find_gl_command([b'glIsRenderbuffer', b'glIsRenderbufferEXT'])

    global glIsShader
    glIsShader = <glIsShader_type> find_gl_command([b'glIsShader'])

    global glIsTexture
    glIsTexture = <glIsTexture_type> find_gl_command([b'glIsTexture'])

    global glIsVertexArray
    glIsVertexArray = <glIsVertexArray_type> find_gl_command([b'glIsVertexArray', b'glIsVertexArrayAPPLE', b'glIsVertexArrayOES'])

    global glLineWidth
    glLineWidth = <glLineWidth_type> find_gl_command([b'glLineWidth'])

    global glLinkProgram
    glLinkProgram = <glLinkProgram_type> find_gl_command([b'glLinkProgram', b'glLinkProgramARB'])

    global glMapBufferRange
    glMapBufferRange = <glMapBufferRange_type> find_gl_command([b'glMapBufferRange', b'glMapBufferRangeEXT'])

    global glPixelStorei
    glPixelStorei = <glPixelStorei_type> find_gl_command([b'glPixelStorei'])

    global glPolygonOffset
    glPolygonOffset = <glPolygonOffset_type> find_gl_command([b'glPolygonOffset'])

    global glReadBuffer
    glReadBuffer = <glReadBuffer_type> find_gl_command([b'glReadBuffer'])

    global glReadPixels
    glReadPixels = <glReadPixels_type> find_gl_command([b'glReadPixels'])

    global glRenderbufferStorage
    glRenderbufferStorage = <glRenderbufferStorage_type> find_gl_command([b'glRenderbufferStorage', b'glRenderbufferStorageEXT'])

    global glRenderbufferStorageMultisample
    glRenderbufferStorageMultisample = <glRenderbufferStorageMultisample_type> find_gl_command([b'glRenderbufferStorageMultisample', b'glRenderbufferStorageMultisampleEXT', b'glRenderbufferStorageMultisampleNV'])

    global glSampleCoverage
    glSampleCoverage = <glSampleCoverage_type> find_gl_command([b'glSampleCoverage', b'glSampleCoverageARB'])

    global glScissor
    glScissor = <glScissor_type> find_gl_command([b'glScissor'])

    global glShaderSource
    glShaderSource = <glShaderSource_type> find_gl_command([b'glShaderSource', b'glShaderSourceARB'])

    global glStencilFunc
    glStencilFunc = <glStencilFunc_type> find_gl_command([b'glStencilFunc'])

    global glStencilFuncSeparate
    glStencilFuncSeparate = <glStencilFuncSeparate_type> find_gl_command([b'glStencilFuncSeparate'])

    global glStencilMask
    glStencilMask = <glStencilMask_type> find_gl_command([b'glStencilMask'])

    global glStencilMaskSeparate
    glStencilMaskSeparate = <glStencilMaskSeparate_type> find_gl_command([b'glStencilMaskSeparate'])

    global glStencilOp
    glStencilOp = <glStencilOp_type> find_gl_command([b'glStencilOp'])

    global glStencilOpSeparate
    glStencilOpSeparate = <glStencilOpSeparate_type> find_gl_command([b'glStencilOpSeparate', b'glStencilOpSeparateATI'])

    global glTexImage2D
    glTexImage2D = <glTexImage2D_type> find_gl_command([b'glTexImage2D'])

    global glTexImage3D
    glTexImage3D = <glTexImage3D_type> find_gl_command([b'glTexImage3D', b'glTexImage3DEXT'])

    global glTexParameterf
    glTexParameterf = <glTexParameterf_type> find_gl_command([b'glTexParameterf'])

    global glTexParameterfv
    glTexParameterfv = <glTexParameterfv_type> find_gl_command([b'glTexParameterfv'])

    global glTexParameteri
    glTexParameteri = <glTexParameteri_type> find_gl_command([b'glTexParameteri'])

    global glTexParameteriv
    glTexParameteriv = <glTexParameteriv_type> find_gl_command([b'glTexParameteriv'])

    global glTexSubImage2D
    glTexSubImage2D = <glTexSubImage2D_type> find_gl_command([b'glTexSubImage2D', b'glTexSubImage2DEXT'])

    global glTexSubImage3D
    glTexSubImage3D = <glTexSubImage3D_type> find_gl_command([b'glTexSubImage3D', b'glTexSubImage3DEXT'])

    global glTransformFeedbackVaryings
    glTransformFeedbackVaryings = <glTransformFeedbackVaryings_type> find_gl_command([b'glTransformFeedbackVaryings', b'glTransformFeedbackVaryingsEXT'])

    global glUniform1f
    glUniform1f = <glUniform1f_type> find_gl_command([b'glUniform1f', b'glUniform1fARB'])

    global glUniform1fv
    glUniform1fv = <glUniform1fv_type> find_gl_command([b'glUniform1fv', b'glUniform1fvARB'])

    global glUniform1i
    glUniform1i = <glUniform1i_type> find_gl_command([b'glUniform1i', b'glUniform1iARB'])

    global glUniform1iv
    glUniform1iv = <glUniform1iv_type> find_gl_command([b'glUniform1iv', b'glUniform1ivARB'])

    global glUniform1ui
    glUniform1ui = <glUniform1ui_type> find_gl_command([b'glUniform1ui', b'glUniform1uiEXT'])

    global glUniform1uiv
    glUniform1uiv = <glUniform1uiv_type> find_gl_command([b'glUniform1uiv', b'glUniform1uivEXT'])

    global glUniform2f
    glUniform2f = <glUniform2f_type> find_gl_command([b'glUniform2f', b'glUniform2fARB'])

    global glUniform2fv
    glUniform2fv = <glUniform2fv_type> find_gl_command([b'glUniform2fv', b'glUniform2fvARB'])

    global glUniform2i
    glUniform2i = <glUniform2i_type> find_gl_command([b'glUniform2i', b'glUniform2iARB'])

    global glUniform2iv
    glUniform2iv = <glUniform2iv_type> find_gl_command([b'glUniform2iv', b'glUniform2ivARB'])

    global glUniform2ui
    glUniform2ui = <glUniform2ui_type> find_gl_command([b'glUniform2ui', b'glUniform2uiEXT'])

    global glUniform2uiv
    glUniform2uiv = <glUniform2uiv_type> find_gl_command([b'glUniform2uiv', b'glUniform2uivEXT'])

    global glUniform3f
    glUniform3f = <glUniform3f_type> find_gl_command([b'glUniform3f', b'glUniform3fARB'])

    global glUniform3fv
    glUniform3fv = <glUniform3fv_type> find_gl_command([b'glUniform3fv', b'glUniform3fvARB'])

    global glUniform3i
    glUniform3i = <glUniform3i_type> find_gl_command([b'glUniform3i', b'glUniform3iARB'])

    global glUniform3iv
    glUniform3iv = <glUniform3iv_type> find_gl_command([b'glUniform3iv', b'glUniform3ivARB'])

    global glUniform3ui
    glUniform3ui = <glUniform3ui_type> find_gl_command([b'glUniform3ui', b'glUniform3uiEXT'])

    global glUniform3uiv
    glUniform3uiv = <glUniform3uiv_type> find_gl_command([b'glUniform3uiv', b'glUniform3uivEXT'])

    global glUniform4f
    glUniform4f = <glUniform4f_type> find_gl_command([b'glUniform4f', b'glUniform4fARB'])

    global glUniform4fv
    glUniform4fv = <glUniform4fv_type> find_gl_command([b'glUniform4fv', b'glUniform4fvARB'])

    global glUniform4i
    glUniform4i = <glUniform4i_type> find_gl_command([b'glUniform4i', b'glUniform4iARB'])

    global glUniform4iv
    glUniform4iv = <glUniform4iv_type> find_gl_command([b'glUniform4iv', b'glUniform4ivARB'])

    global glUniform4ui
    glUniform4ui = <glUniform4ui_type> find_gl_command([b'glUniform4ui', b'glUniform4uiEXT'])

    global glUniform4uiv
    glUniform4uiv = <glUniform4uiv_type> find_gl_command([b'glUniform4uiv', b'glUniform4uivEXT'])

    global glUniformMatrix2fv
    glUniformMatrix2fv = <glUniformMatrix2fv_type> find_gl_command([b'glUniformMatrix2fv', b'glUniformMatrix2fvARB'])

    global glUniformMatrix2x3fv
    glUniformMatrix2x3fv = <glUniformMatrix2x3fv_type> find_gl_command([b'glUniformMatrix2x3fv', b'glUniformMatrix2x3fvNV'])

    global glUniformMatrix2x4fv
    glUniformMatrix2x4fv = <glUniformMatrix2x4fv_type> find_gl_command([b'glUniformMatrix2x4fv', b'glUniformMatrix2x4fvNV'])

    global glUniformMatrix3fv
    glUniformMatrix3fv = <glUniformMatrix3fv_type> find_gl_command([b'glUniformMatrix3fv', b'glUniformMatrix3fvARB'])

    global glUniformMatrix3x2fv
    glUniformMatrix3x2fv = <glUniformMatrix3x2fv_type> find_gl_command([b'glUniformMatrix3x2fv', b'glUniformMatrix3x2fvNV'])

    global glUniformMatrix3x4fv
    glUniformMatrix3x4fv = <glUniformMatrix3x4fv_type> find_gl_command([b'glUniformMatrix3x4fv', b'glUniformMatrix3x4fvNV'])

    global glUniformMatrix4fv
    glUniformMatrix4fv = <glUniformMatrix4fv_type> find_gl_command([b'glUniformMatrix4fv', b'glUniformMatrix4fvARB'])

    global glUniformMatrix4x2fv
    glUniformMatrix4x2fv = <glUniformMatrix4x2fv_type> find_gl_command([b'glUniformMatrix4x2fv', b'glUniformMatrix4x2fvNV'])

    global glUniformMatrix4x3fv
    glUniformMatrix4x3fv = <glUniformMatrix4x3fv_type> find_gl_command([b'glUniformMatrix4x3fv', b'glUniformMatrix4x3fvNV'])

    global glUnmapBuffer
    glUnmapBuffer = <glUnmapBuffer_type> find_gl_command([b'glUnmapBuffer', b'glUnmapBufferARB', b'glUnmapBufferOES'])

    global glUseProgram
    glUseProgram = <glUseProgram_type> find_gl_command([b'glUseProgram', b'glUseProgramObjectARB'])

    global glValidateProgram
    glValidateProgram = <glValidateProgram_type> find_gl_command([b'glValidateProgram', b'glValidateProgramARB'])

    global glVertexAttrib1f
    glVertexAttrib1f = <glVertexAttrib1f_type> find_gl_command([b'glVertexAttrib1f', b'glVertexAttrib1fARB', b'glVertexAttrib1fNV'])

    global glVertexAttrib1fv
    glVertexAttrib1fv = <glVertexAttrib1fv_type> find_gl_command([b'glVertexAttrib1fv', b'glVertexAttrib1fvARB', b'glVertexAttrib1fvNV'])

    global glVertexAttrib2f
    glVertexAttrib2f = <glVertexAttrib2f_type> find_gl_command([b'glVertexAttrib2f', b'glVertexAttrib2fARB', b'glVertexAttrib2fNV'])

    global glVertexAttrib2fv
    glVertexAttrib2fv = <glVertexAttrib2fv_type> find_gl_command([b'glVertexAttrib2fv', b'glVertexAttrib2fvARB', b'glVertexAttrib2fvNV'])

    global glVertexAttrib3f
    glVertexAttrib3f = <glVertexAttrib3f_type> find_gl_command([b'glVertexAttrib3f', b'glVertexAttrib3fARB', b'glVertexAttrib3fNV'])

    global glVertexAttrib3fv
    glVertexAttrib3fv = <glVertexAttrib3fv_type> find_gl_command([b'glVertexAttrib3fv', b'glVertexAttrib3fvARB', b'glVertexAttrib3fvNV'])

    global glVertexAttrib4f
    glVertexAttrib4f = <glVertexAttrib4f_type> find_gl_command([b'glVertexAttrib4f', b'glVertexAttrib4fARB', b'glVertexAttrib4fNV'])

    global glVertexAttrib4fv
    glVertexAttrib4fv = <glVertexAttrib4fv_type> find_gl_command([b'glVertexAttrib4fv', b'glVertexAttrib4fvARB', b'glVertexAttrib4fvNV'])

    global glVertexAttribI4i
    glVertexAttribI4i = <glVertexAttribI4i_type> find_gl_command([b'glVertexAttribI4i', b'glVertexAttribI4iEXT'])

    global glVertexAttribI4iv
    glVertexAttribI4iv = <glVertexAttribI4iv_type> find_gl_command([b'glVertexAttribI4iv', b'glVertexAttribI4ivEXT'])

    global glVertexAttribI4ui
    glVertexAttribI4ui = <glVertexAttribI4ui_type> find_gl_command([b'glVertexAttribI4ui', b'glVertexAttribI4uiEXT'])

    global glVertexAttribI4uiv
    glVertexAttribI4uiv = <glVertexAttribI4uiv_type> find_gl_command([b'glVertexAttribI4uiv', b'glVertexAttribI4uivEXT'])

    global glVertexAttribIPointer
    glVertexAttribIPointer = <glVertexAttribIPointer_type> find_gl_command([b'glVertexAttribIPointer', b'glVertexAttribIPointerEXT'])

    global glVertexAttribPointer
    glVertexAttribPointer = <glVertexAttribPointer_type> find_gl_command([b'glVertexAttribPointer', b'glVertexAttribPointerARB'])

    global glViewport
    glViewport = <glViewport_type> find_gl_command([b'glViewport'])



from __future__ import division, absolute_import, with_statement, print_function, unicode_literals

import os
import renpy
import textwrap


def shaders(incdir="source/inc"):

    def p(s):
        print(s, file=outf)

    def indented(s):
        s = textwrap.dedent(s)
        for i in s.split("\n"):
            print("    " + i, file=outf)

    with open(os.path.join(incdir, "shadersource"), "w") as outf:

        parts = list(renpy.gl2.gl2shadercache.shader_part.values())

        def priority(sp):
            if sp.vertex_parts:
                return sp.vertex_parts[0][0], sp.name
            else:
                return sp.fragment_parts[0][0], sp.name

        parts.sort(key=priority)

        for sp in parts:

            if sp.name == "renpy.ftl":
                continue

            header = "{} (priority {})".format(sp.name, priority(sp)[0])

            p(header)
            p("^" * len(header))

            if sp.raw_variables:
                p("Variables::")

                s = textwrap.dedent(sp.raw_variables)

                indented(sp.raw_variables)

            if sp.vertex_functions or sp.fragment_functions:
                raise Exception("Can't doc functions yet.")

            for _, s in sp.vertex_parts:
                p("Vertex shader::")

                indented(s)

            for _, s in sp.fragment_parts:
                p("Fragment shader::")

                indented(s)




# Copyright 2004-2022 Tom Rothamel <pytom@bishoujo.us>
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

# This file contains code for formatting tracebacks.

from __future__ import division, absolute_import, with_statement, print_function, unicode_literals
from renpy.compat import PY2, basestring, bchr, bord, chr, open, pystr, range, round, str, tobytes, unicode # *



import traceback
import sys
import io
import platform
import linecache
import time
import os

import renpy

FSENCODING = sys.getfilesystemencoding() or "utf-8"


def write_traceback_list(out, l):
    """
    Given the traceback list, writes it to out as unicode.
    """

    ul = [ ]

    for filename, line, what, text in l:

        # Filename is either unicode or fsecoded bytes.
        if isinstance(filename, bytes):
            filename = filename.decode(FSENCODING)

        # Line is a number.

        # Assume what is in a unicode encoding, since it is either python,
        # or comes from inside Ren'Py.

        if isinstance(text, bytes):
            text = text.decode("utf-8")

        ul.append((filename, line, what, text))

    for t in traceback.format_list(ul):
        out.write(t)


def traceback_list(tb):
    """
    Given `tb`, returns a list of (filename, line_number, function, line_text)
    tuples.
    """

    l = [ ]

    while tb:
        frame = tb.tb_frame
        line_number = tb.tb_lineno
        code = frame.f_code
        filename = code.co_filename
        name = code.co_name

        tb = tb.tb_next

        if ('self' in frame.f_locals) and (not renpy.config.raw_tracebacks):
            obj = frame.f_locals['self']

            last = (tb is None)

            try:
                report = obj.report_traceback(name, last)

                if report is not None:
                    l.extend(report)
                    continue
            except Exception:
                pass

        l.append((filename, line_number, name, None))

    rv = [ ]

    for filename, line_number, name, line in l:
        if line is None:
            try:
                line = linecache.getline(filename, line_number)
            except Exception:
                line = ''

        rv.append((filename, line_number, name, line))

    return rv


def filter_traceback_list(tl):
    """
    Returns the subset of `tl` that originates in creator-written files, as
    opposed to those portions that come from Ren'Py itself.
    """

    rv = [ ]

    for t in tl:
        filename = t[0]
        if filename.endswith(".rpy") and not filename.replace("\\", "/").startswith("common/"):
            rv.append(t)

    return rv


def open_error_file(fn, mode):
    """
    Opens an error/log/file. Returns the open file, and the filename that
    was opened.
    """

    try:
        new_fn = os.path.join(renpy.config.logdir, fn) # type: ignore
        f = open(new_fn, mode)
        return f, new_fn
    except Exception:
        pass

    try:
        f = open(fn, mode)
        return f, fn
    except Exception:
        pass

    import tempfile

    new_fn = os.path.join(tempfile.gettempdir(), "renpy-" + fn)
    return open(new_fn, mode), new_fn


def report_exception(e, editor=True):
    """
    Reports an exception by writing it to standard error and
    traceback.txt. If `editor` is True, opens the traceback
    up in a text editor.

    Returns a three-item tuple, with the first item being
    a simplified traceback, the second being a full traceback,
    and the third being the traceback filename.
    """

    # Note: Doki Doki Literature club calls this as ("Words...", False).
    # For what it's worth.

    import codecs

    type, _value, tb = sys.exc_info() # @ReservedAssignment

    # Return values - which can be displayed to the user.
    simple = io.StringIO()
    full = io.StringIO()

    full_tl = traceback_list(tb)
    simple_tl = filter_traceback_list(full_tl)

    print(str(renpy.game.exception_info), file=simple)
    write_traceback_list(simple, simple_tl)
    print(type.__name__ + ":", end=' ', file=simple)
    print(str(e), file=simple)

    print("Full traceback:", file=full)
    write_traceback_list(full, full_tl)
    print(type.__name__ + ":", end=' ', file=full)
    print(str(e), file=full)

    # Write to stdout/stderr.
    try:
        sys.stdout.write("\n")
        sys.stdout.write(full.getvalue())
        sys.stdout.write("\n")
        sys.stdout.write(simple.getvalue())
    except Exception:
        pass

    print('', file=full)

    try:
        print(str(platform.platform()), str(platform.machine()), file=full)
        print(renpy.version, file=full)
        print(renpy.config.name + " " + renpy.config.version, file=full)
        print(str(time.ctime()), file=full)
    except Exception:
        pass

    simple = simple.getvalue()
    full = full.getvalue()

    # Inside of the file, which may not be openable.
    try:

        f, traceback_fn = open_error_file("traceback.txt", "w")

        with f:
            f.write("\ufeff") # BOM

            print("I'm sorry, but an uncaught exception occurred.", file=f)
            print('', file=f)

            f.write(simple)

            print('', file=f)
            print("-- Full Traceback ------------------------------------------------------------", file=f)
            print('', file=f)

            f.write(full)

        try:
            renpy.util.expose_file(traceback_fn)
        except Exception:
            pass

        try:
            if editor and ((renpy.game.args.command == "run") or (renpy.game.args.errors_in_editor)): # type: ignore
                renpy.exports.launch_editor([ traceback_fn ], 1, transient=True)
        except Exception:
            pass

    except Exception:
        traceback_fn = os.path.join(renpy.config.basedir, "traceback.txt") # type: ignore

    return simple, full, traceback_fn



# Copyright 2004-2022 Tom Rothamel <pytom@bishoujo.us>
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import division, absolute_import, with_statement, print_function, unicode_literals
from renpy.compat import PY2, basestring, bchr, bord, chr, open, pystr, range, round, str, tobytes, unicode # *

from typing import Optional

import renpy
import time
import collections

import datetime

# Profiling ####################################################################

profile_log = renpy.log.open("profile_screen", developer=True, append=False, flush=False)

# A map from screen name to ScreenProfile object.
profile = { }


class ScreenProfile(renpy.object.Object):
    """
    :doc: profile_screen
    :name: renpy.profile_screen

    """

    def __init__(self, name, predict=False, show=False, update=False, request=False, time=False, debug=False, const=False):
        """
        Requests screen profiling for the screen named `name`, which
        must be a string.

        Apart from `name`, all arguments must be supplied as keyword
        arguments. This function takes three groups of arguments.


        The first group of arguments determines when profiling occurs.

        `predict`
            If true, profiling occurs when the screen is being predicted.

        `show`
            If true, profiling occurs when the screen is first shown.

        `update`
            If true, profiling occurs when the screen is updated.

        `request`
            If true, profiling occurs when requested by pressing F8.

        The second group of arguments controls what profiling output is
        produced when profiling occurs.

        `time`
            If true, Ren'Py will log the amount of time it takes to evaluate
            the screen.

        `debug`
            If true, Ren'Py will log information as to how screens are
            evaluated, including:

            * Which displayables Ren'Py considers constant.
            * Which arguments, if any, needed to be evaluated.
            * Which displayables were reused.

            Producing and saving this debug information takes a noticeable
            amount of time, and so the `time` output should not be considered
            reliable if `debug` is set.

        The last group of arguments controls what output is produced once
        per Ren'Py run.

        `const`
            Displays the variables in the screen that are marked as const and
            not-const.

        All profiling output will be logged to profile_screen.txt in the game
        directory.
        """

        self.predict = predict
        self.show = show
        self.update = update
        self.request = request

        self.time = time
        self.debug = debug

        self.const = const

        if name is not None:
            if isinstance(name, basestring):
                name = tuple(name.split())
                profile[name] = self


def get_profile(name):
    """
    Returns the profile object for the screen with `name`, or a default
    profile object if none exists.

    `name`
        A string or tuple.
    """

    if isinstance(name, basestring):
        name = tuple(name.split())

    if name in profile:
        return profile[name]
    else:
        return ScreenProfile(None)

# Cache ########################################################################


# A map from screen name to a list of ScreenCache objects. We ensure the cache
# does not exceed config.screen_cache_size for each screen.
predict_cache = collections.defaultdict(list)


class ScreenCache(object):
    """
    Represents an entry in the screen cache. Upon creation, puts itself into
    the screen cache.
    """

    def __init__(self, screen, args, kwargs, cache):

        if screen.ast is None:
            return

        self.args = args
        self.kwargs = kwargs
        self.cache = cache

        pc = predict_cache[screen]

        pc.append(self)

        if len(pc) > renpy.config.screen_cache_size:
            pc.pop(0)


cache_put = ScreenCache


def cache_get(screen, args, kwargs):
    """
    Returns the cache to use when `screen` is accessed with `args` and
    `kwargs`.
    """

    if screen.ast is None:
        return { }

    pc = predict_cache[screen]

    if not pc:
        return { }

    for sc in pc:

        # Reuse w/ same arguments.
        if sc.args == args and sc.kwargs == kwargs:
            pc.remove(sc)
            break
    else:

        # Reuse the oldest.
        sc = pc.pop(0)

    return sc.cache

# Screens #####################################################################


class Screen(renpy.object.Object):
    """
    A screen is a collection of widgets that are displayed together.
    This class stores information about the screen.
    """

    sensitive = "True"
    roll_forward = None

    def __init__(self,
                 name,
                 function,
                 modal="False",
                 zorder="0",
                 tag=None,
                 predict=None,
                 variant=None,
                 parameters=False,
                 location=None,
                 layer="screens",
                 sensitive="True",
                 roll_forward=None):

        # The name of this screen.
        if isinstance(name, basestring):
            name = tuple(name.split())

        self.name = name

        if (variant is None) or isinstance(variant, basestring):
            variant = [ variant ]

        for v in variant:
            screens[name[0], v] = self
            screens_by_name[name[0]][v] = self

        # A function that can be called to display the screen.
        self.function = function

        # If this is a SL2 screen, the SLScreen node at the root of this
        # screen.
        if isinstance(function, renpy.sl2.slast.SLScreen): # @UndefinedVariable
            self.ast = function
        else:
            self.ast = None

        # Expression: Are we modal? (A modal screen ignores screens under it.)
        self.modal = modal

        # Expression: Our zorder.
        self.zorder = zorder

        # The tag associated with the screen.
        self.tag = tag or name[0]

        # Can this screen be predicted?
        if predict is None:
            predict = renpy.config.predict_screens

        self.predict = predict

        # True if this screen takes parameters via _args and _kwargs.
        self.parameters = parameters

        # The location (filename, linenumber) of this screen.
        self.location = location

        # The layer the screen will be shown on.
        self.layer = layer

        # Is this screen sensitive? An expression.
        self.sensitive = sensitive

        # Does the screen participate in roll_forward when used with the
        # call screen statement? True for yes, False for no, None for
        # config.call_screen_roll_forward.
        self.roll_forward = roll_forward

        global prepared
        global analyzed

        prepared = False
        analyzed = False


# Phases we can be in.
PREDICT = 0 # Predicting the screen before it is shown.
SHOW = 1 # Showing the screen for the first time.
UPDATE = 2 # Showing the screen for the second and later times.
HIDE = 3 # After the screen has been hid with "hide screen" (or the end of call screen).
OLD = 4 # A copy of the screen in the old side of a transition.

phase_name = [
    "PREDICT",
    "SHOW",
    "UPDATE",
    "HIDE",
    "OLD",
    ]


class ScreenDisplayable(renpy.display.layout.Container):
    """
    A screen is a collection of widgets that are displayed together. This
    class is responsible for managing the display of a screen.
    """

    nosave = [
        'screen',
        'child',
        'children',
        'transforms',
        'widgets',
        'base_widgets',
        'old_widgets',
        'hidden_widgets',
        'old_transforms',
        'cache',
        'miss_cache',
        'profile',
        'phase',
        'use_cache' ]

    restarting = False
    hiding = False
    transient = False

    def after_setstate(self):
        self.screen = get_screen_variant(self.screen_name[0])
        self.child = None
        self.children = [ ]
        self.transforms = { }
        self.widgets = { }
        self.base_widgets = { }
        self.old_widgets = None
        self.old_transforms = None
        self.old_transfers = False
        self.hidden_widgets = { }
        self.cache = { }
        self.phase = UPDATE
        self.use_cache = { }
        self.miss_cache = { }

        self.profile = profile.get(self.screen_name, None)

    def __init__(self, screen, tag, layer, widget_properties={}, scope={}, transient=False, **properties):

        super(ScreenDisplayable, self).__init__(**properties)

        # Stash the properties, so we can re-create the screen.
        self.properties = properties

        # The screen, and it's name. (The name is used to look up the
        # screen on save.)
        self.screen = screen
        self.screen_name = screen.name

        self._location = self.screen.location

        # The profile object that determines when we profile.
        self.profile = profile.get(self.screen_name, None)

        # The tag and layer screen was displayed with.
        self.tag = tag
        self.layer = layer

        # The scope associated with this statement. This is passed in
        # as keyword arguments to the displayable.
        self.scope = renpy.revertable.RevertableDict(scope)

        # The child associated with this screen.
        self.child = None # type: renpy.display.layout.MultiBox|None

        # Widget properties given to this screen the last time it was
        # shown.
        self.widget_properties = widget_properties

        # A map from name to the widget with that name.
        self.widgets = { }

        # Same, but to the widget without considering _main.
        self.base_widgets = { }

        # The persistent cache.
        self.cache = { }

        if tag and layer:
            old_screen = get_screen(tag, layer)
        else:
            old_screen = None

        # A map from name to the transform with that name. (This is
        # taken from the old version of the screen, if it exists.
        if old_screen is not None:
            self.transforms = old_screen.transforms
        else:
            self.transforms = { }

        # A map from a (screen name, id) pair to cache. This is for use
        # statements with the id parameter.
        if old_screen is not None:
            self.use_cache = old_screen.use_cache
        else:
            self.use_cache = { }

        # A version of the cache that's used when we have a screen that is
        # being displayed with the same tag with a cached copy of the screen
        # we want to display.
        self.miss_cache = { }

        # What widgets and transforms were the last time this screen was
        # updated. Used to communicate with the ui module, and only
        # valid during an update - not used at other times.
        self.old_widgets = None
        self.old_transforms = None

        # Should we transfer data from the old_screen? This becomes
        # true once this screen finishes updating for the first time,
        # and also while we're using something.
        self.old_transfers = (old_screen and old_screen.screen_name == self.screen_name)

        # The current transform event, and the last transform event to
        # be processed.
        self.current_transform_event = None

        # A dict-set of widgets (by id) that have been hidden from us.
        self.hidden_widgets = { }

        # Are we restarting or hiding?
        self.restarting = False
        self.hiding = False

        # Is this a transient screen?
        self.transient = transient

        # Modal and zorder.
        self.modal = renpy.python.py_eval(self.screen.modal, locals=self.scope)
        self.zorder = renpy.python.py_eval(self.screen.zorder, locals=self.scope)

        # The lifecycle phase we are in - one of PREDICT, SHOW, UPDATE, or HIDE.
        self.phase = PREDICT

    @property
    def name(self):
        return " ".join(self.screen_name)

    def _repr_info(self):
        return self.name

    def visit(self):
        return [ self.child ]

    def visit_all(self, callback, seen=None):
        callback(self)

        try:
            push_current_screen(self)
            if self.child is not None:
                self.child.visit_all(callback, seen=None)
        finally:
            pop_current_screen()

    def per_interact(self):
        renpy.display.render.redraw(self, 0)
        self.update()

    def set_transform_event(self, event):
        super(ScreenDisplayable, self).set_transform_event(event)
        self.current_transform_event = event

    def find_focusable(self, callback, focus_name):

        hiding = (self.phase == OLD) or (self.phase == HIDE)

        if self.modal and not callable(self.modal):
            renpy.display.focus.mark_modal()

        try:
            push_current_screen(self)

            if self.child and not hiding:
                self.child.find_focusable(callback, focus_name)
        finally:
            pop_current_screen()


    def copy(self):
        rv = ScreenDisplayable(self.screen, self.tag, self.layer, self.widget_properties, self.scope, **self.properties)
        rv.transforms = self.transforms.copy()
        rv.widgets = self.widgets.copy()
        rv.base_widgets = self.widgets.copy()
        rv.old_transfers = True
        rv.child = self.child

        return rv

    def _handles_event(self, event):
        if self.child is None:

            if self.transient:
                return False

            self.update()

        return self.child._handles_event(event)

    def _hide(self, st, at, kind):

        if self.phase == HIDE:
            hid = self
        else:

            if (self.child is not None) and (not self.child._handles_event(kind)):
                return None

            updated_screens.discard(self)
            self.update()

            if self.screen is None:
                return None

            if self.child is None:
                return None

            if not self.child._handles_event(kind):
                return None

            if self.screen.ast is not None:
                self.screen.ast.copy_on_change(self.cache.get(0, {}))

            hid = self.copy()

            for i in self.child.children:
                i.set_transform_event(kind)

        hid.phase = HIDE

        rv = None

        old_child = hid.child

        if not isinstance(old_child, renpy.display.layout.MultiBox):
            return None

        renpy.ui.detached()
        hid.child = renpy.ui.default_fixed(focus="_screen_" + "_".join(self.screen_name))
        hid.children = [ hid.child ]
        renpy.ui.close()

        for d in old_child.children:
            c = d._hide(st, at, kind)

            if c is not None:
                renpy.display.render.redraw(c, 0)
                hid.child.add(c)

                rv = hid

        if hid is not None:
            renpy.display.render.redraw(hid, 0)

        return rv

    def _in_current_store(self):

        if self.screen is None:
            return self

        if self.child is None:
            return self

        if not renpy.config.transition_screens:
            return self

        if self.screen.ast is not None:
            self.screen.ast.copy_on_change(self.cache.get(0, {}))

        rv = self.copy()
        rv.phase = OLD
        rv.child = self.child._in_current_store()

        return rv

    def update(self):

        if self in updated_screens:
            return

        updated_screens.add(self)

        if self.screen is None:
            self.child = renpy.display.layout.MultiBox(layout='fixed')
            return { }

        # Do not update if restarting or hiding.
        if self.restarting or (self.phase == HIDE) or (self.phase == OLD):
            if not self.child:
                self.child = renpy.display.layout.MultiBox(layout='fixed')

            return self.widgets

        profile = False
        debug = False

        if self.profile:

            if self.phase == UPDATE and self.profile.update:
                profile = True
            elif self.phase == SHOW and self.profile.show:
                profile = True
            elif self.phase == PREDICT and self.profile.predict:
                profile = True

            if renpy.display.interface.profile_once and self.profile.request:
                profile = True

            if profile:
                profile_log.write("%s %s %s",
                                  phase_name[self.phase],
                                  " ".join(self.screen_name),
                                  datetime.datetime.now().strftime("%H:%M:%S.%f"))

                start = time.time()

                if self.profile.debug:
                    debug = True

        # Cycle widgets and transforms.
        self.old_widgets = self.widgets
        self.old_transforms = self.transforms
        self.widgets = { }
        self.base_widgets = { }
        self.transforms = { }

        push_current_screen(self)

        old_ui_screen = renpy.ui.screen
        renpy.ui.screen = self

        # The name of the root screen of this screen.
        NAME = 0

        old_cache = self.cache.get(NAME, None)

        # Evaluate the screen.
        try:

            renpy.ui.detached()
            self.child = renpy.ui.default_fixed(focus="_screen_" + "_".join(self.screen_name))
            self.children = [ self.child ]

            self.scope["_scope"] = self.scope
            self.scope["_name"] = NAME
            self.scope["_debug"] = debug

            self.screen.function(**self.scope)

            renpy.ui.close()

        finally:
            # Safe removal as to not reraise another exception and lose the last one
            self.scope.pop("_scope", None)

            renpy.ui.screen = old_ui_screen
            pop_current_screen()

        # Finish up.
        self.old_widgets = None
        self.old_transforms = None
        self.old_transfers = True

        if self.miss_cache:
            self.miss_cache.clear()

        # Deal with the case where the screen version changes.
        if (self.cache.get(NAME, None) is not old_cache) and (self.current_transform_event is None) and (self.phase == UPDATE):
            self.current_transform_event = "update"

        if self.current_transform_event:

            for i in self.child.children:
                i.set_transform_event(self.current_transform_event)

            self.current_transform_event = None

        if profile:
            end = time.time()

            if self.profile.time:
                profile_log.write("* %.2f ms", 1000 * (end - start)) # type: ignore

            if self.profile.debug:
                profile_log.write("\n")

        return self.widgets

    def render(self, w, h, st, at):

        if not self.child:
            self.update()

        if self.phase == SHOW:
            self.phase = UPDATE

        try:
            push_current_screen(self)
            child = renpy.display.render.render(self.child, w, h, st, at)
        finally:
            pop_current_screen()

        rv = renpy.display.render.Render(w, h)
        rv.focus_screen = self

        hiding = (self.phase == OLD) or (self.phase == HIDE)

        if self.screen is None:
            sensitive = False
        else:
            sensitive = renpy.python.py_eval(self.screen.sensitive, locals=self.scope)

        rv.blit(child, (0, 0), focus=sensitive and not hiding, main=not hiding)
        rv.modal = self.modal and not hiding

        return rv

    def get_placement(self):
        if not self.child:
            self.update()

        return self.child.get_placement()

    def event(self, ev, x, y, st):

        if (self.phase == OLD) or (self.phase == HIDE):
            return

        if not self.screen:
            return None

        if not renpy.python.py_eval(self.screen.sensitive, locals=self.scope):
            ev = renpy.display.interface.time_event

        try:
            push_current_screen(self)

            rv = self.child.event(ev, x, y, st)
        finally:
            pop_current_screen()

        if rv is not None:
            return rv

        if renpy.display.layout.check_modal(self.modal, ev, x, y, None, None):
            raise renpy.display.layout.IgnoreLayers()

    def get_phase_name(self):
        return phase_name[self.phase]

    def _tts(self):
        if (self.phase == OLD) or (self.phase == HIDE):
            return ""

        if self.modal:
            return renpy.display.tts.TTSDone(self._tts_common())
        else:
            return self._tts_common()


# The name of the screen that is currently being displayed, or
# None if no screen is being currently displayed.
_current_screen = None # type: renpy.display.screen.ScreenDisplayable|None

if 0 == 1:
    _current_screen = renpy.display.screen.ScreenDisplayable() # type: ignore # fake out typing.
else:
    _current_screen = None

# The stack of old current screens.
current_screen_stack = [ ]


def push_current_screen(screen):
    global _current_screen
    current_screen_stack.append(_current_screen)
    _current_screen = screen


def pop_current_screen():
    global _current_screen
    _current_screen = current_screen_stack.pop()


# A map from (screen_name, variant) tuples to screen.
screens = { }

# A map from screen name to map from variant to screen.
screens_by_name = collections.defaultdict(dict)

# The screens that were updated during the current interaction.
updated_screens = set()


def get_screen_variant(name, candidates=None):
    """
    Get a variant screen object for `name`.

    `candidates`
        A list of candidate variants.
    """

    if candidates is None:
        candidates = renpy.config.variants

    for i in candidates:
        rv = screens.get((name, i), None)
        if rv is not None:
            return rv

    return None


def get_all_screen_variants(name):
    """
    Gets all variants of the screen with `name`.

    Returns a list of (`variant`, `screen`) tuples, in no particular
    order.
    """

    if isinstance(name, basestring):
        name = tuple(name.split())

    name = name[0]

    if name not in screens_by_name:
        return [ ]

    return list(screens_by_name[name].items())


# Have all screens been analyzed?
analyzed = False

# Have the screens been prepared?
prepared = False

# Caches for sort_screens.
sorted_screens = [ ]
screens_at_sort = { }

# The list of screens that participate in a use cycle.
use_cycle = [ ]


def sort_screens():
    """
    Produces a list of SL2 screens in topologically sorted order.
    """

    global use_cycle
    global sorted_screens
    global screens_at_sort

    if screens_at_sort == screens:
        return sorted_screens

    # For each screen, the set of screens it uses.
    depends = collections.defaultdict(set)

    # For each screen, the set of screens that use it.
    reverse = collections.defaultdict(set)

    names = { i[0] for i in screens }

    for k, v in screens.items():

        name = k[0]

        # Ensure name exists.
        depends[name]

        if not v.ast:
            continue

        def callback(uses):

            if uses not in names:
                return

            depends[name].add(uses)
            reverse[uses].add(name)

        v.ast.used_screens(callback)

    rv = [ ]

    workset = { k for k, v in depends.items() if not len(v) }

    while workset:
        name = workset.pop()
        rv.append(name)

        for i in reverse[name]:
            d = depends[i]
            d.remove(name)

            if not d:
                workset.add(i)

        del reverse[name]

    # Store the use cycle for later reporting.
    use_cycle = list(reverse.keys())
    use_cycle.sort()

    sorted_screens = rv
    screens_at_sort = dict(screens)

    return rv


def sorted_variants():
    """
    Produces a list of screen variants in topological order.
    """

    rv = [ ]

    for name in sort_screens():
        rv.extend(screens_by_name[name].values())

    return rv


def analyze_screens():
    """
    Analyzes all screens.
    """

    global analyzed

    if analyzed:
        return

    for s in sorted_variants():
        if s.ast is None:
            continue

        s.ast.analyze_screen()

    analyzed = True


def prepare_screens():
    """
    Prepares all screens for use.
    """

    global prepared

    if prepared:
        return

    predict_cache.clear()

    old_predicting = renpy.display.predict.predicting
    renpy.display.predict.predicting = True

    try:

        if not analyzed:
            analyze_screens()

        for s in sorted_variants():
            if s.ast is None:
                continue

            s.ast.unprepare_screen()
            s.ast.prepare_screen()

        prepared = True

    finally:
        renpy.display.predict.predicting = old_predicting

    if renpy.config.developer and use_cycle:
        raise Exception("The following screens use each other in a loop: " + ", ".join(use_cycle) + ". This is not allowed.")


def define_screen(*args, **kwargs):
    """
    :undocumented:
    :args: (name, function, modal="False", zorder="0", tag=None, variant=None)

    Defines a screen with `name`, which should be a string.

    `function`
        The function that is called to display the screen. The
        function is called with the screen scope as keyword
        arguments. It should ignore additional keyword arguments.

        The function should call the ui functions to add things to the
        screen.

    `modal`
        A string that, when evaluated, determines of the created
        screen should be modal. A modal screen prevents screens
        underneath it from receiving input events.

    `zorder`
        A string that, when evaluated, should be an integer. The integer
        controls the order in which screens are displayed. A screen
        with a greater zorder number is displayed above screens with a
        lesser zorder number.

    `tag`
        The tag associated with this screen. When the screen is shown,
        it replaces any other screen with the same tag. The tag
        defaults to the name of the screen.

    `predict`
        If true, this screen can be loaded for image prediction. If false,
        it can't. Defaults to true.

    `variant`
        String. Gives the variant of the screen to use.

    """

    Screen(*args, **kwargs)


def get_screen_layer(name):
    """
    Returns the layer that the screen with `name` is part of.
    """

    if not isinstance(name, basestring):
        name = name[0]

    screen = get_screen_variant(name)

    if screen is None:
        return "screens"
    else:
        return screen.layer


def get_screen(name, layer=None):
    """
    :doc: screens

    Returns the ScreenDisplayable with the given `name` on `layer`. `name`
    is first interpreted as a tag name, and then as a screen name. If the
    screen is not showing, returns None.

    This can also take a list of names, in which case the first screen
    that is showing is returned.

    This function can be used to check if a screen is showing::

        if renpy.get_screen("say"):
            text "The say screen is showing."
        else:
            text "The say screen is hidden."

    The ScreenDisplayable objects returned by this function have the
    following documented fields:

    .. attribute:: ScreenDisplayable.layer

        The layer the screen is being displayed on.

    .. attribute:: ScreenDisplayable.name

        The name of the screen.

    .. attribute:: ScreenDisplayable.zorder

        The zorder the screen is being displayed at.
    """

    if layer is None:
        layer = get_screen_layer(name)

    if isinstance(name, basestring):
        name = (name,)

    sl = renpy.exports.scene_lists()

    for tag in name:

        sd = sl.get_displayable_by_tag(layer, tag)
        if sd is not None:
            return sd

    for tag in name:

        sd = sl.get_displayable_by_name(layer, (tag,))
        if sd is not None:
            return sd

    return None


def has_screen(name):
    """
    Returns true if a screen with the given name exists.
    """

    if not isinstance(name, tuple):
        name = tuple(name.split())

    if not name:
        return False

    if get_screen_variant(name[0]):
        return True
    else:
        return False


def get_screen_roll_forward(screen_name):
    """
    Given a screeen name, determines if roll forward is enable for the
    screen.
    """

    name = screen_name

    if not isinstance(name, tuple):
        name = tuple(name.split())

    screen = get_screen_variant(name[0])

    if screen is None:
        raise Exception("Screen %s is not known.\n" % (name[0],))

    return screen.roll_forward


def show_screen(_screen_name, *_args, **kwargs):
    """
    :doc: screens

    The programmatic equivalent of the show screen statement.

    Shows the named screen. This takes the following keyword arguments:

    `_screen_name`
        The name of the screen to show.
    `_layer`
        The layer to show the screen on.
    `_zorder`
        The zorder to show the screen on. If not specified, defaults to
        the zorder associated with the screen. It that's not specified,
        it is 0 by default.
    `_tag`
        The tag to show the screen with. If not specified, defaults to
        the tag associated with the screen. It that's not specified,
        defaults to the name of the screen.
    `_widget_properties`
        A map from the id of a widget to a property name -> property
        value map. When a widget with that id is shown by the screen,
        the specified properties are added to it.
    `_transient`
        If true, the screen will be automatically hidden at the end of
        the current interaction.

    Non-keyword arguments, and keyword arguments that do not begin with
    an underscore, are passed to the screen.
    """

    _layer = kwargs.pop("_layer", None)
    _tag = kwargs.pop("_tag", None)
    _widget_properties = kwargs.pop("_widget_properties", {})
    _transient = kwargs.pop("_transient", False)
    _zorder = kwargs.pop("_zorder", None)

    name = _screen_name

    if not isinstance(name, tuple):
        name = tuple(name.split())

    screen = get_screen_variant(name[0])

    if screen is None:
        raise Exception("Screen %s is not known.\n" % (name[0],))

    if _layer is None:
        _layer = get_screen_layer(name)

    if _tag is None:
        _tag = screen.tag

    scope = { }

    if screen.parameters:
        scope["_kwargs" ] = kwargs
        scope["_args"] = _args
    else:
        scope.update(kwargs)

    d = ScreenDisplayable(screen, _tag, _layer, _widget_properties, scope, transient=_transient)

    if _zorder is None:
        _zorder = d.zorder

    old_d = get_screen(_tag, _layer)

    if old_d and old_d.cache:
        d.cache = old_d.cache
        d.miss_cache = cache_get(screen, _args, kwargs)
        d.phase = UPDATE
    else:
        d.cache = cache_get(screen, _args, kwargs)
        d.phase = SHOW

    sls = renpy.display.core.scene_lists()

    sls.add(_layer, d, _tag, zorder=_zorder, transient=_transient, keep_st=True, name=name)


def predict_screen(_screen_name, *_args, **kwargs):
    """
    Predicts the displayables that make up the given screen.

    `_screen_name`
        The name of the  screen to show.
    `_widget_properties`
        A map from the id of a widget to a property name -> property
        value map. When a widget with that id is shown by the screen,
        the specified properties are added to it.

    Keyword arguments not beginning with underscore (_) are used to
    initialize the screen's scope.
    """

    _layer = kwargs.pop("_layer", None)
    _tag = kwargs.pop("_tag", None)
    _widget_properties = kwargs.pop("_widget_properties", {})
    _transient = kwargs.pop("_transient", False)

    name = _screen_name

    if renpy.config.debug_prediction:
        renpy.display.ic_log.write("Predict screen %s", name)

    if not isinstance(name, tuple):
        name = tuple(name.split())

    screen = get_screen_variant(name[0])

    if screen is None:
        return

    if not screen.predict:
        return

    if _layer is None:
        _layer = get_screen_layer(name)

    scope = { }
    scope["_scope"] = scope

    if screen.parameters:
        scope["_kwargs" ] = kwargs
        scope["_args"] = _args
    else:
        scope.update(kwargs)

    try:

        d = ScreenDisplayable(screen, None, None, _widget_properties, scope)
        d.cache = cache_get(screen, _args, kwargs)
        d.update()
        cache_put(screen, _args, kwargs, d.cache)

        renpy.display.predict.displayable(d)

    except Exception:
        if renpy.config.debug_prediction:
            import traceback

            print("While predicting screen", _screen_name)
            traceback.print_exc()
            print()

    finally:
        scope.pop("_scope", None)

    renpy.ui.reset()


def hide_screen(tag, layer=None):
    """
    :doc: screens

    The programmatic equivalent of the hide screen statement.

    Hides the screen with `tag` on `layer`.
    """

    if layer is None:
        layer = get_screen_layer((tag,))

    screen = get_screen(tag, layer)

    if screen is not None:
        renpy.exports.hide(screen.tag, layer=layer)


def use_screen(_screen_name, *_args, **kwargs):

    _name = kwargs.pop("_name", ())
    _scope = kwargs.pop("_scope", { })

    name = _screen_name

    if not isinstance(name, tuple):
        name = tuple(name.split())

    screen = get_screen_variant(name[0])

    if screen is None:
        raise Exception("Screen %r is not known." % (name,))

    old_transfers = _current_screen.old_transfers
    _current_screen.old_transfers = True

    if screen.parameters:
        scope = { }
        scope["_kwargs"] = kwargs
        scope["_args"] = _args
    else:
        scope = _scope.copy()
        scope.update(kwargs)

    scope["_scope"] = scope
    scope["_name"] = (_name, name)

    try:
        screen.function(**scope)
    finally:
        scope.pop("_scope", None)

    _current_screen.old_transfers = old_transfers


def current_screen(): # type: () -> ScreenDisplayable|None
    """
    :doc: screens
    :name: renpy.current_screen

    Returns the ScreenDisplayable corresponding to the screen currently being
    updated, rendered, or processed.

    See :func:`get_screen` for documented fields on ScreenDisplayable.
    """

    return _current_screen


def get_displayable(screen, id, layer=None, base=False): # @ReservedAssignment
    """
    :doc: screens
    :name: renpy.get_displayable

    From the `screen` on `layer`, returns the displayable with
    `id`. Returns None if the screen doesn't exist, or there is no
    widget with that id on the screen.
    """

    if isinstance(screen, ScreenDisplayable):
        screen = screen.screen_name

    if screen is None:
        screen = current_screen()
    else:
        if layer is None:
            layer = get_screen_layer(screen)

        screen = get_screen(screen, layer)

    if not isinstance(screen, ScreenDisplayable):
        return None

    if screen.child is None:
        screen.update()

    if base:
        rv = screen.base_widgets.get(id, None)
    else:
        rv = screen.widgets.get(id, None)

    return rv


get_widget = get_displayable


def get_displayable_properties(id, screen=None, layer=None): # @ReservedAssignment
    """
    :doc: screens
    :name: renpy.get_displayable_properties

    Returns the properties for the displayable with `id` in the `screen`
    on `layer`. If `screen` is None, returns the properties for the
    current screen. This can be used from Python or property code inside
    a screen.

    Note that this returns a dictionary containing the widget properties,
    and so to get an individual property, the dictionary must be accessed.
    """

    if screen is None:
        s = current_screen()
    else:
        if layer is None:
            layer = get_screen_layer(screen)

        s = get_screen(screen, layer)

    if s is None:
        return { }

    rv = s.widget_properties.get(id, None)

    if rv is None:
        rv = { }

    return rv


get_widget_properties = get_displayable_properties


def before_restart():
    """
    This is called before Ren'Py restarts to put the screens into restart
    mode, which prevents crashes due to variables being used that are no
    longer defined.
    """

    for k, layer in renpy.display.interface.old_scene.items():
        if k is None:
            continue

        for i in layer.children:
            if isinstance(i, ScreenDisplayable):
                i.restarting = True


def show_overlay_screens(suppress_overlay):
    """
    Called from interact to show or hide the overlay screens.
    """

    show = not suppress_overlay

    if renpy.store._overlay_screens is None:
        show = show
    elif renpy.store._overlay_screens is True:
        show = True
    else:
        show = False

    if show:

        for i in renpy.config.overlay_screens:
            if get_screen(i) is None:
                show_screen(i)

    else:

        for i in renpy.config.overlay_screens:
            if get_screen(i) is not None:
                hide_screen(i)

    for i in renpy.config.always_shown_screens:
        if get_screen(i) is None:
            show_screen(i)



def per_frame():
    """
    Called from interact once per frame to invalidate screens we want to
    update once per frame.
    """

    for i in renpy.config.per_frame_screens:
        s = get_screen(i)

        if s is None:
            continue

        updated_screens.discard(s)
        renpy.display.render.invalidate(s)
        s.update()



# Copyright 2004-2022 Tom Rothamel <pytom@bishoujo.us>
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import division, absolute_import, with_statement, print_function, unicode_literals
from renpy.compat import PY2, basestring, bchr, bord, chr, open, pystr, range, round, str, tobytes, unicode # *



import collections
import renpy
import os

# A map from filename to position, target label pairs.
missing = collections.defaultdict(list)


def report_missing(target, filename, position):
    """
    Reports that the call statement ending at `position` in `filename`
    is missing a from clause.
    """

    missing[filename].append((position, target))


# Labels that we've created while running add_from.
new_labels = set()


def generate_label(target):
    """
    Generate a reasonable and unique new label for a call to `target`.
    """

    target = target.replace(".", "_")

    n = 0

    while True:
        if n:
            label = "_call_{}_{}".format(target, n)
        else:
            label = "_call_{}".format(target)

        if not renpy.exports.has_label(label) and not (label in new_labels):
            break

        n += 1

    new_labels.add(label)
    return label


def process_file(fn):
    """
    Adds missing from clauses to `fn`.
    """

    if not os.path.exists(fn):
        return

    edits = missing[fn]
    edits.sort()

    with open(fn, "rb") as f:
        data = f.read().decode("utf-8")

    # How much of the input has been consumed.
    consumed = 0

    # The output.
    output = u""

    for position, target in edits:
        output += data[consumed:position]
        consumed = position

        output += " from {}".format(generate_label(target))

    output += data[consumed:]

    with open(fn + ".new", "wb") as f:
        f.write(output.encode("utf-8"))

    try:
        os.unlink(fn + ".bak")
    except Exception:
        pass

    os.rename(fn, fn + ".bak")
    os.rename(fn + ".new", fn)


def add_from():

    renpy.arguments.takes_no_arguments("Adds from clauses to call statements that are missing them.")

    for fn in missing:
        if fn.startswith(renpy.config.gamedir):
            process_file(fn)

    return False


renpy.arguments.register_command("add_from", add_from)



# Copyright 2004-2022 Tom Rothamel <pytom@bishoujo.us>
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import division, absolute_import, with_statement, print_function, unicode_literals
from renpy.compat import PY2, basestring, bchr, bord, chr, open, pystr, range, round, str, tobytes, unicode # *



import renpy
import json


def extract_strings_core(language, destination, merge=False, force=False):

    if (not force) and (language not in renpy.game.script.translator.strings):  # @UndefinedVariable
        raise Exception("Language %r does not have any translations." % language)

    st = renpy.game.script.translator.strings[language]  # @UndefinedVariable

    result = { }

    if merge:
        with open(destination, "r") as f:
            result.update(json.load(f, encoding="utf-8"))

    for k, v in st.translations.items():
        if v and v != k:
            result[k] = v

    with open(destination, "w") as f:
        json.dump(result, f, ensure_ascii=True)


def extract_strings():
    """
    The extract strings command.
    """

    ap = renpy.arguments.ArgumentParser(description="Extracts translated strings.")
    ap.add_argument("language", help="The language to extract translated strings from.")
    ap.add_argument("destination", help="The json file to store the translated strings into.")
    ap.add_argument("--merge", help="If given, the current contents of the file are preserved, and new contents are merged into the file.", action="store_true")
    ap.add_argument("--force", help="If given, noting happens if the language does not exist.", action="store_true")

    args = ap.parse_args()

    language = args.language

    if language == 'None':
        language = None

    extract_strings_core(language, args.destination, args.merge, args.force)

    return False


renpy.arguments.register_command("extract_strings", extract_strings)



# Copyright 2004-2022 Tom Rothamel <pytom@bishoujo.us>
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

# Other text-related things.

from __future__ import division, absolute_import, with_statement, print_function, unicode_literals
from renpy.compat import PY2, basestring, bchr, bord, chr, open, pystr, range, round, str, tobytes, unicode # *



import renpy

from renpy.text.textsupport import DISPLAYABLE, PARAGRAPH, TAG
import renpy.text.textsupport as textsupport

# A list of text tags, mapping from the text tag prefix to if it
# requires a closing tag.
text_tags = dict(
    alpha=True,
    alt=True,
    art=True,
    done=False,
    image=False,
    p=False,
    w=False,
    fast=False,
    b=True,
    i=True,
    u=True,
    a=True,
    plain=True,
    font=True,
    color=True,
    outlinecolor=True,
    size=True,
    noalt=True,
    nw=False,
    s=True,
    rt=True,
    rb=True,
    k=True,
    cps=True,
    space=False,
    vspace=False
    )

text_tags[""] = True


# This checks the text tags in a string to be sure they are all matched, and
# properly nested. It returns an error message, or None if the line is okay.
def check_text_tags(s):
    """
    :doc: lint

    Checks the text tags in s for correctness. Returns an error string if there is
    an error, or None if there is no error.
    """

    all_tags = dict(text_tags)

    custom_tags = renpy.config.custom_text_tags
    if custom_tags:
        all_tags.update(custom_tags)

    self_closing_custom_tags = renpy.config.self_closing_custom_text_tags
    if self_closing_custom_tags:
        all_tags.update(dict.fromkeys(self_closing_custom_tags, False))

    try:
        tokens = textsupport.tokenize(str(s))
    except Exception as e:
        return e.args[0]

    tag_stack = [ ]

    for type, text in tokens: # @ReservedAssignment
        if type != TAG:
            continue

        if text[0] == "#":
            continue

        # Strip off arguments for tags.
        if text.find('=') != -1:
            text = text[:text.find('=')]

        # Closing tag.
        if text and text[0] == '/':
            if not tag_stack:
                return "Close text tag '%s' does not match an open text tag." % text

            if tag_stack[-1] != text[1:]:
                return "Close text tag '%s' does not match open text tag '%s'." % (text, tag_stack[-1])

            tag_stack.pop()
            continue

        if text not in all_tags:
            return "Text tag '%s' is not known." % text

        if all_tags[text]:
            tag_stack.append(text)

    if tag_stack:
        return "One or more text tags were left open at the end of the string: " + ", ".join([ "'" + i + "'" for i in tag_stack])

    return None


def filter_text_tags(s, allow=None, deny=None):
    """
    :doc: text_utility

    Returns a copy of `s` with the text tags filtered. Exactly one of the `allow` and `deny` keyword
    arguments must be given.

    `allow`
        A set of tags that are allowed. If a tag is not in this list, it is removed.

    `deny`
        A set of tags that are denied. If a tag is not in this list, it is kept in the string.
    """

    if (allow is None) and (deny is None):
        raise Exception("Only one of the allow and deny keyword arguments should be given to filter_text_tags.")

    if (allow is not None) and (deny is not None):
        raise Exception("Only one of the allow and deny keyword arguments should be given to filter_text_tags.")

    tokens = textsupport.tokenize(str(s))

    rv = [ ]

    for tokentype, text in tokens:

        if tokentype == PARAGRAPH:
            rv.append("\n")
        elif tokentype == TAG:
            kind = text.partition("=")[0]

            if kind and (kind[0] == "/"):
                kind = kind[1:]

            if allow is not None:
                if kind in allow:
                    rv.append("{" + text + "}")
            else:
                if kind not in deny:
                    rv.append("{" + text + "}")
        else:
            rv.append(text.replace("{", "{{"))

    return "".join(rv)


def filter_alt_text(s):
    """
    Returns a copy of `s` with the contents of text tags that shouldn't be in
    alt text filtered. This returns just the text to say, with no text tags
    at all in it.
    """

    tokens = textsupport.tokenize(str(s))

    if renpy.config.custom_text_tags or renpy.config.self_closing_custom_text_tags or (renpy.config.replace_text is not None):
        tokens = renpy.text.text.Text.apply_custom_tags(tokens)

    rv = [ ]

    active = set()

    for tokentype, text in tokens:

        if tokentype == PARAGRAPH:
            rv.append("\n")
        elif tokentype == TAG:
            kind = text.partition("=")[0]

            if kind.startswith("/"):
                kind = kind[1:]
                end = True
            else:
                end = False

            if kind in renpy.config.tts_filter_tags:
                if end:
                    active.discard(kind)
                else:
                    active.add(kind)
        elif tokentype == DISPLAYABLE:
            rv.append(text._tts())
        else:
            if not active:
                rv.append(text)

    return "".join(rv)


class ParameterizedText(object):
    """
    :name: ParameterizedText
    :doc: text

    This is a displayable that can be shown with an additional string
    parameter, which then shows that string as if it was an image.
    This is usually used as part of the pre-defined ``text`` image.

    For example, one can do::

        show text "Hello, World" at truecenter
        with dissolve
        pause 1
        hide text
        with dissolve

    You can use ParameterizedText directly to define similar images with
    different style properties. For example, one can write::

        image top_text = ParameterizedText(xalign=0.5, yalign=0.0)

        label start:
            show top_text "This text is shown at the center-top of the screen"
    """

    def __init__(self, style='default', **properties):
        self.style = style
        self.properties = properties

    _duplicatable = True

    def _duplicate(self, args):

        if args.lint:
            return renpy.text.text.Text("", style=self.style, **self.properties)

        if len(args.args) == 0:
            raise Exception("'%s' takes a single string parameter." % ' '.join(args.name))

        param = "".join(args.args)
        string = renpy.python.py_eval(param)

        return renpy.text.text.Text(string, style=self.style, **self.properties)


def textwrap(s, width=78, asian=False):
    """
    Wraps the unicode string `s`, and returns a list of strings.

    `width`
        The number of half-width characters that fit on a line.
    `asian`
        True if we should make ambiguous width characters full-width, as is
        done in Asian encodings.
    """

    import unicodedata

    glyphs = [ ]

    for c in str(s):

        eaw = unicodedata.east_asian_width(c)

        if (eaw == "F") or (eaw == "W"):
            gwidth = 20
        elif (eaw == "A"):
            if asian:
                gwidth = 20
            else:
                gwidth = 10
        else:
            gwidth = 10

        g = textsupport.Glyph()
        g.character = ord(c)
        g.ascent = 10
        g.line_spacing = 10
        g.width = gwidth
        g.advance = gwidth

        glyphs.append(g)

    textsupport.annotate_unicode(glyphs, False, 2)
    renpy.text.texwrap.linebreak_tex(glyphs, width * 10, width * 10, False)
    return textsupport.linebreak_list(glyphs)
